(*
   module Main (.ml file only)
   Deals with processing the command line, reading files, building and
   connecting lexers and parsers.
*)

open Lolevel.Error
open Hilevel.Printing
open Types
open Engine

let byline = "NYU Lambda Seminar Untyped interpreter"
let search_path = ref [""]
let terms = ref []
let rec arg_defs = [
  ("-I", Arg.String (fun dir -> search_path := dir::!search_path), "Append a directory to the search path");
  ("-v", Arg.Unit (fun () -> begin print_endline byline; print_endline ("version %VERSION%" ^ version); exit 0 end), "Print version info");
  ("-e", Arg.String (fun str -> terms := str::!terms), "Supply term(s) to be reduced/evaluated")
]

let parse_args parse_file parse_string =
  let infile = ref (None : string option) in
  let handle_one str =
    match !infile with
    | Some _ -> error "You must specify exactly one input file."
    | None -> infile := Some str in
  let arg_defs' = [
    ("-", Arg.Unit (fun () -> handle_one "-"), "");
    ("--", Arg.Rest handle_one, "");
    ("-help", Arg.Unit (fun () -> raise (Arg.Bad "unknown option `-help`")), "")
  ] @ arg_defs in
  begin
    Arg.parse arg_defs' handle_one byline;
    match !infile with
    | Some s ->
      (match !terms with
      | [] -> parse_file s
      | _ -> error "Cannot handle both -e terms and a source file.")
    | None ->
      (match !terms with
      | [] -> error "You must specify an input file (use `-` for stdin)."
      | terms -> parse_string (String.concat ";" (List.rev terms)))
  end

let open_file infile =
  let rec trynext = function
    | [] -> error ("Could not find " ^ infile)
    | dir::dirs ->
        let path = if dir = "" then infile else (Filename.concat dir infile) in
        try open_in path
        with Sys_error m -> trynext dirs in
  if infile = "-" then stdin else trynext (List.rev !search_path)

let parse_file infile =
  let chan = open_file infile in
  let lexbuf = Lexer.from_channel (if infile = "-" then "/dev/stdin" else infile) chan in
  let (result : env -> toplevel_syntax list * env) =
    try Parser.total Lexer.main lexbuf
    with Parsing.Parse_error ->
      error_with (Lexer.info lexbuf) "Syntax error." in
  begin
    Parsing.clear_parser();
    close_in chan;
    result
  end

let parse_string str =
  let lexbuf = Lexer.from_string str in
  let (result : env -> toplevel_syntax list * env) =
    try Parser.total Lexer.main lexbuf
    with Parsing.Parse_error ->
      error_with (Lexer.info lexbuf) "Syntax error." in
  begin
    Parsing.clear_parser();
    result
  end

let rec process_toplevel env = function
  | Term(_,t) ->
      let res = interpret t env in
      begin
        print_result true res;
        force_newline();
        env
      end
  | BindTerm(_,x,t) ->
      let res = interpret t env in
      begin
        print x;
        print " ";
        print_binding (TopBind res);
        force_newline();
        shift x (TopBind res) env
      end
  | BindRec(_) ->
      failwith "letrec not implemented"
  | BindSymbol(_,x) ->
      begin
        print "symbol ";
        print x;
        force_newline();
        shift x SymbolBind env
      end

let process_string str =
  let res = Printexc.print (fun () ->
    try
      let (tops,env') = parse_string str empty in
      (* we discard the env' because it has been filled with _unevaluated_ toplevel bindings *)
      let aux env c = begin
        obox0();
        let env' = process_toplevel env c in
        flush();
        env'
      end in
      let _ = List.fold_left aux empty tops in
      0
    with Exit x -> x) () in
  let () = flush() in
  res

let main () =
  let res = Printexc.print (fun () ->
    try
      let do_parse = parse_args parse_file parse_string in
      let (tops,env') = do_parse empty in
      (* we discard the env' because it has been filled with _unevaluated_ toplevel bindings *)
      let aux env c = begin
        obox0();
        let env' = process_toplevel env c in
        flush();
        env'
      end in
      let _ = List.fold_left aux empty tops in
      0
    with Exit x -> x) () in
    exit res

let () = Format.set_max_boxes 1000
let () = Format.set_margin 67
let () = if not !Sys.interactive then main ()

